from ... import utils


def docs(header, reduction=None):
    def arguments():
        return r"""
Arguments
---------
data: Tuple[torch.Tensor, torch.Tensor]
    Tuple containing `outputs` from neural network and `targets` (ground truths).
    `outputs` should be of shape :math:`(N, *)` and contain `logits` or `probabilities`.
    `targets` should be of shape :math:`(N, *)` as well and contain `boolean` values
    (or integers from set :math:`{0, 1}`).

        """

    def wrapper(klass):
        docstring = r"""{}

Works for both logits and probabilities of `output`.

If `output` is tensor after `sigmoid` activation
user should change `threshold` to `0.5` for
correct results (default `0.0` corresponds to unnormalized probability a.k.a logits).

Parameters
----------
threshold : float, optional
    Threshold above which prediction is considered to be positive.
    Default: `0.0`
        """.format(
            header
        )
        if reduction is not None:
            docstring += utils.docs_general.reduction_parameter(reduction)

        klass.forward.__doc__ = arguments() + utils.docs_general.returned(reduction)
        klass.__doc__ = docstring
        return klass

    return wrapper



import abc
import collections
import typing

import torch

from ... import _base, functional
from . import utils

###############################################################################
#
#                           COMMON BASE CLASSES
#
###############################################################################


class _ReductionSum(_base.Operation):
    def __init__(self, reduction=torch.sum):
        super().__init__()
        self.reduction = reduction

    @abc.abstractmethod
    def forward(self, data):
        pass


class _ReductionMean(_base.Operation):
    def __init__(self, reduction=torch.mean):
        super().__init__()
        self.reduction = reduction

    @abc.abstractmethod
    def forward(self, data):
        pass


###############################################################################
#
#                          CONCRETE IMPLEMENTATIONS
#
###############################################################################


###############################################################################
#
#                           MEAN DEFAULT REDUCTION
#
###############################################################################


@utils.multiclass.docs(
    header="""Calculate accuracy score between `output` and `target`.""",
    reduction="mean",
)
class Accuracy(_ReductionMean):
    def forward(self, data):
        return functional.metrics.classification.multiclass.accuracy(
            *data, self.reduction
        )


@utils.multiclass.docs(
    header="""Calculate Jaccard score between `output` and `target`.""",
    reduction="mean",
)
class Jaccard(_ReductionMean):
    def forward(self, data):
        return functional.metrics.classification.multiclass.jaccard(
            *data, self.reduction
        )


###############################################################################
#
#                           SUM DEFAULT REDUCTION
#
###############################################################################


# Basic cases
@utils.multiclass.docs(
    header="""Number of false positives between `output` and `target`.""",
    reduction="sum",
)
class TruePositive(_ReductionSum):
    def forward(self, data):
        return functional.metrics.classification.multiclass.true_positive(
            *data, self.reduction
        )


@utils.multiclass.docs(
    header="""Number of false positives between `output` and `target`.""",
    reduction="sum",
)
class FalsePositive(_ReductionSum):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_positive(
            *data, self.reduction
        )


@utils.multiclass.docs(
    header="""Number of true negatives between `output` and `target`.""",
    reduction="sum",
)
class TrueNegative(_ReductionSum):
    def forward(self, data):
        return functional.metrics.classification.multiclass.true_negative(
            *data, self.reduction
        )


@utils.multiclass.docs(
    header="""Number of false negatives between `output` and `target`.""",
    reduction="sum",
)
class FalseNegative(_ReductionSum):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_negative(
            *data, self.reduction
        )


@utils.multiclass.docs(
    header="""Confusion matrix between `output` and `target`.""", reduction="sum",
)
class ConfusionMatrix(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.confusion_matrix(*data,)


###############################################################################
#
#                               NO REDUCTION
#
###############################################################################


@utils.multiclass.docs(header="""Recall between `output` and `target`.""",)
class Recall(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.recall(*data)


@utils.multiclass.docs(header="""Specificity between `output` and `target`.""",)
class Specificity(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.specificity(*data)


@utils.multiclass.docs(header="""Precision between `output` and `target`.""",)
class Precision(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.precision(*data)


@utils.multiclass.docs(
    header="""Negative predictive value between `output` and `target`.""",
)
class NegativePredictiveValue(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.negative_predictive_value(
            *data
        )


@utils.multiclass.docs(header="""False negative rate between `output` and `target`.""",)
class FalseNegativeRate(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_negative_rate(*data)


@utils.multiclass.docs(header="""False positive rate between `output` and `target`.""",)
class FalsePositiveRate(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_positive_rate(*data)


@utils.multiclass.docs(
    header="""False discovery rate between `output` and `target`.""",
)
class FalseDiscoveryRate(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_discovery_rate(*data)


@utils.multiclass.docs(header="""False omission rate between `output` and `target`.""",)
class FalseOmissionRate(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.false_omission_rate(*data)


@utils.multiclass.docs(
    header="""Critical success index between `output` and `target`.""",
)
class CriticalSuccessIndex(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.critical_success_index(
            *data
        )


@utils.multiclass.docs(
    header="""Critical success index between `output` and `target`.""",
)
class BalancedAccuracy(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.balanced_accuracy(*data)


@utils.multiclass.docs(header="""F1 score between `output` and `target`.""",)
class F1(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.f1(*data)


@utils.multiclass.docs(
    header="""Matthews correlation coefficient between `output` and `target`.""",
)
class MatthewsCorrelationCoefficient(_base.Operation):
    def forward(self, data):
        return functional.metrics.classification.multiclass.matthews_correlation_coefficient(
            *data
        )


###############################################################################
#
#                               OTHER METRICS
#
###############################################################################


class FBeta(_base.Operation):
    r"""Get f-beta score between `outputs` and `targets`.

    Works for both logits and probabilities of `output` out of the box.

    Parameters
    ----------
    beta: float
        Beta coefficient of `f-beta` score.

    Arguments
    ---------
    data: Tuple[torch.Tensor, torch.Tensor]
        Tuple containing `outputs` from neural network and `targets` (ground truths).
        `outputs` should be of shape :math:`(N, *, C-1)`, where :math:`C` is the number of classes.
        Should contain `logits` (unnormalized probabilities) or `probabilities` after
        `softmax` activation or similar.
        `targets` should be of shape :math:`(N, *)` and contain integers in the range
        :math:`[0, C-1]`


    Returns
    -------
    torch.Tensor
        Scalar `tensor`

    """

    def __init__(self, beta: float):
        super().__init__()
        self.beta = beta

    def forward(self, data):
        return functional.metrics.classification.multiclass.f_beta(*data, self.beta)


class TopK(_base.Operation):
    r"""Get top-k accuracy score between `outputs` and `targets`.

    Works for both logits and probabilities of `output` out of the box.

    Parameters
    ----------
    k: int
        How many top results should be chosen.
    reduction: Callable, optional
        One argument callable getting `torch.Tensor` and returning `torch.Tensor`.
        Default: `torch.sum` (sum of all elements, user can use `torchtraining.savers.Sum`
        to get sum across iterations/epochs).

    Arguments
    ---------
    data: Tuple[torch.Tensor, torch.Tensor]
        Tuple containing `outputs` from neural network and `targets` (ground truths).
        `outputs` should be of shape :math:`(N, *, C-1)`, where :math:`C` is the number of classes.
        Should contain `logits` (unnormalized probabilities) or `probabilities` after
        `softmax` activation or similar.
        `targets` should be of shape :math:`(N, *)` and contain integers in the range
        :math:`[0, C-1]`

    Returns
    -------
    torch.Tensor
        If `reduction` is left as default mean is taken and single value returned.
        Otherwise whatever `reduction` returns.

    """

    def __init__(self, k: int, reduction=torch.mean):
        super().__init__()
        self.k = k
        self.reduction = reduction

    def forward(self, data):
        return functional.metrics.classification.multiclass.topk(*data, self.reduction)



"""This module provides standard PyTorch operations (like `backward`)
in functional manner.

.. note::

    **IMPORTANT**: This module is used almost all the time
    so be sure to understand how it works.


It allows users to perform training on single `step` for both training and evaluation
using PyTorch's optimizer, backward or zeroing gradient, for example::


    class Step(tt.steps.Step):
        def forward(self, module, sample):
            # Your forward step here
            ...
            return loss, predictions

    training = (
        Step(criterion, gradient=True, device=device)
        ** tt.Select(loss=0)
        ** tt.pytorch.ZeroGrad(network)
        ** tt.pytorch.Backward()
        ** tt.pytorch.Optimize(optimizer)
        ** tt.pytorch.Detach()
    )

    evaluation = (
        Step(criterion, gradient=False, device=device)
        ** tt.Select(predictions=1)
        ** tt.callbacks.Log(writer, "Predicted")
    )

Some other operations are also simplified (e.g. gradient accumulation),
see `torchtraining.callbacks.Optimize`

"""

import torch

from ._base import Operation


class Detach(Operation):
    """Returns a new Tensor, detached from the current graph.

    .. note::

        **IMPORTANT**: This operation should be used before accumulating
        values after `iteration` in order not to grow backpropagation
        graph.

    Returns
    -------
    torch.Tensor
        Detached tensor

    """

    def forward(self, data):
        """
        Arguments
        ---------
        torch.Tensor
            Tensor to be detached (new Tensor is returned).

        """
        return data.detach()


class Schedule(Operation):
    """Run single step of given scheduler.

    Usually placed after each `step` or `iteration` (depending on provided
    scheduler instance).

    Returns
    -------
    torch.Tensor
        Value passed to function initially

    Parameters
    ----------
    scheduler : torch.optim.lr_scheduler._LRScheduler
        Instance of scheduler-like object with interface aligned with
        `torch.optim.lr_scheduler._LRScheduler` base class
    use_data : bool
        Whether input data should be used when stepping scheduler.
    """

    def __init__(self, scheduler, use_data: bool = False):
        super().__init__()
        self.scheduler = scheduler
        self.use_data = use_data

    def forward(self, data):
        """
        Arguments
        ---------
        torch.Tensor
            Tensor which is optionally used to step scheduler.

        """
        if self.use_data:
            self.scheduler.step(data)
        else:
            self.scheduler.step()

        return data


class Backward(Operation):
    """Run backpropagation on output tensor.

    Parameters
    ----------
    scaler : torch.cuda.amp.GradScaler, optional
        Gradient scaler used for automatic mixed precision mode.
    accumulate: int, optional
        Divide loss by ``accumulate`` if gradient accumulation is used.
        This approach averages gradient from multiple batches.
        Default: `1` (no accumulation)
    gradient : torch.Tensor, optional
        Tensor used as initial value to backpropagation. If unspecified,
        uses `torch.tensor([1.0])` as default value (just like `tensor.backward()` call).

    Returns
    -------
    torch.Tensor
        Tensor after backward (possibly scaled by `accumulate`)

    """

    def __init__(self, scaler=None, accumulate: int = 1, gradient: torch.Tensor = None):
        super().__init__()
        self.scaler = scaler
        self.accumulate = accumulate
        self.gradient = gradient

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor on which `backward` will be run (possibly accumulated).
            Usually `loss` value

        """
        output = data / self.accumulate
        if self.scaler is not None:
            output = self.scaler.scale(output)
        if self.gradient is not None:
            output.backward(self.gradient)
        else:
            output.backward()

        return output


class Optimize(Operation):
    """Perform optimization step on `parameters` stored by `optimizer`.

    Currently specifying `closure` and `scaler` is mutually exclusive.

    Parameters
    ----------
    optimizer: torch.optim.Optimizer
        Instance of optimizer-like object with interface aligned with
        `torch.optim.Optimizer`.
    accumulate: int, optional
        Divide loss by ``accumulate`` if gradient accumulation is used.
        This approach averages gradient from multiple batches.
        Default: `1` (no accumulation)
    closure : Callable, optional
        A closure that reevaluates the model and returns the loss.
        Optional for most optimizers. Default: `None`
    scaler : torch.cuda.amp.GradScaler, optional
        Gradient scaler used for automatic mixed precision mode.
        Default: `None`
    *args
        Arguments passed to either `scaler.step` (if specified) or `optimizer.step`
    **kwargs
        Keyword arguments passed to either `scaler.step` (if specified) or `optimizer.step`

    Returns
    -------
    Any
        Anything passed to `forward`.


    """

    def __init__(
        self, optimizer, accumulate: int = 1, closure=None, scaler=None, *args, **kwargs
    ):
        super().__init__()
        self.optimizer = optimizer
        self.accumulate = accumulate

        if scaler is not None and closure is not None:
            raise ValueError("Closure use with scaler is not currently supported.")

        self.scaler = scaler
        self.closure = closure
        self.args = args
        self.kwargs = kwargs

        self._counter = -1

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as it does not influence this operation.

        """
        self._counter += 1
        if self._counter % self.accumulate:
            if self.scaler is not None:
                self.scaler.step(self.optimizer, *self.args, **self.kwargs)
            else:
                self.optimizer.step(self.closure, *self.args, **self.kwargs)

        return data


class ZeroGrad(Operation):
    """Zero model or optimizer gradients.

    Function `zero_grad()` will be run on the provided object.
    Usually called after every `step` (or after multiple steps, see `accumulate`
    argument).

    Parameters
    ----------
    obj : torch.optim.Optimizer | torch.nn.Module
        Instance of object to zero gradient on.
    accumulate : int
        Accumulate gradient for specified number of iterations before zero-ing out
        gradient.

    Returns
    -------
    Any
        Anything passed to `forward`.


    """

    def __init__(self, obj, accumulate: int = 1):
        super().__init__()
        self.obj = obj
        self.accumulate = accumulate
        self._counter = -1

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as it does not influence this operation.

        """
        self._counter += 1
        if self._counter % self.accumulate:
            self.obj.zero_grad()
        return data


class UpdateGradScaler(Operation):
    """Update gradient scaler used with automatic mixed precision.

    Parameters
    ----------
    scaler : torch.cuda.amp.GradScaler
        Gradient scaler used for automatic mixed precision mode.

    Returns
    -------
    Any
        Anything passed to `forward`.


    """

    def __init__(self, scaler):
        super().__init__()
        self.scaler = scaler

    def forward(self, data):
        """
        Arguments
        ---------
        data: Any
            Anything as it does not influence this operation.

        """
        self.scaler.update()
        return data



import typing

import torch

from ... import utils
from . import utils as binary_utils


@utils.docs
def accuracy(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.mean,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return reduction((output == target).float())


@utils.docs
def jaccard(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.mean,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    union = (output | target).sum(axis=-1)
    intersection = (target & output).sum(axis=-1)
    empty = union <= 0
    union[empty] = 1
    intersection[empty] = 1

    return reduction(intersection.float() / union)


# Basic cases


@utils.docs
def true_positive(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return reduction((output & target).float())


@utils.docs
def false_positive(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return reduction((output & ~target).float())


@utils.docs
def true_negative(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return reduction((~output & ~target).float())


@utils.docs
def false_negative(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return reduction((~output & target).float())


# Confusion matrix


@utils.docs
def confusion_matrix(
    output: torch.Tensor,
    target: torch.Tensor,
    threshold: float = 0.0,
    reduction: typing.Callable[[torch.Tensor], torch.Tensor] = torch.sum,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    tp = reduction((output & target).float())
    fp = reduction((output & ~target).float())
    tn = reduction((~output & ~target).float())
    fn = reduction((~output & target).float())

    return torch.tensor([tp, fn, fp, tn]).reshape(2, 2, -1).squeeze()


# Rate metrics


@utils.docs
def recall(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return (output & target).sum().float() / target.sum()


@utils.docs
def specificity(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    inverse_target = ~target

    return (~output & inverse_target).sum().float() / inverse_target.sum()


@utils.docs
def precision(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return (output & target).sum().float() / output.sum()


@utils.docs
def negative_predictive_value(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    inverse_output = ~output

    return (inverse_output & ~target).sum().float() / inverse_output.sum()


@utils.docs
def false_negative_rate(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return (~output & target).sum().float() / target.sum()


@utils.docs
def false_positive_rate(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    inverse_target = ~target

    return (output & inverse_target).sum().float() / inverse_target.sum()


@utils.docs
def false_discovery_rate(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    return (output & ~target).sum().float() / output.sum()


@utils.docs
def false_omission_rate(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    inverse_output = ~output

    return (inverse_output & target).sum().float() / inverse_output.sum()


@utils.docs
def critical_success_index(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    tp = (output & target).sum().float()

    return tp / tp + (output != target).sum()


@utils.docs
def balanced_accuracy(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)
    inverse_target = ~target

    return (
        (output & target).sum().float() / target.sum()
        + (~output & inverse_target).sum().float() / inverse_target.sum()
    ) / 2


@utils.docs
def f1(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    tp = 2 * (output & target).sum().float()

    return tp / (tp + (output != target).sum())


@utils.docs
def f_beta(
    output: torch.Tensor, target: torch.Tensor, beta: float, threshold: float = 0.0,
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    tp = (1 + beta) ** 2 * (output & target).sum().float()

    return tp / (tp + (beta ** 2) * (output != target).sum())


@utils.docs
def matthews_correlation_coefficient(
    output: torch.Tensor, target: torch.Tensor, threshold: float = 0.0
) -> torch.Tensor:
    binary_utils.binary.check(output, target)
    output, target = binary_utils.binary.threshold(output, target, threshold)

    inverse_output = ~output
    inverse_target = ~target

    dimensions = tuple(range(len(inverse_output.shape)))

    tp = (output & target).float().sum(dim=dimensions[1:])
    tn = (inverse_output & inverse_target).float().sum(dim=dimensions[1:])
    fp = (output & inverse_target).float().sum(dim=dimensions[1:])
    fn = (inverse_output & target).float().sum(dim=dimensions[1:])

    numerator = torch.dot(tp, tn) - torch.dot(fp, fn)
    denominator = (
        output.sum() * target.sum() * inverse_target.sum() * inverse_output.sum()
    )

    if denominator == 0.0:
        return numerator
    return numerator / denominator



def docs(function):
    """Get module and function name and point users to non-functional counter-part.

    Gets functions name, splits on "_" and transforms to UpperCase class name.

    Redirects users to documentation of non-functional counterpart of `function`.
    """
    # removes functional from module
    module = "torchtraining" + function.__module__[21:]
    function.__doc__ = """See `{}.{}` for details.""".format(
        module,
        "".join([subname.capitalize() for subname in function.__name__.split("_")]),
    )
    return function



import itertools

import torch

import pytest
import torchtraining as tt


@pytest.mark.parametrize(
    "inputs,gamma",
    itertools.product(
        list(
            itertools.permutations(
                (torch.randn(64), torch.randn(64, 3, 28, 28), torch.randn(64, 64)), r=2,
            )
        ),
        (0.0, 1.0, 0.2, 0.5),
    ),
)
def test_mixup(inputs, gamma):
    inputs, labels = inputs
    expected = gamma == 1

    copy_inputs, copy_labels = inputs.clone(), labels.clone()
    mixed_inputs, mixed_labels = tt.functional.inputs.mixup(inputs, labels, gamma)
    assert (copy_inputs == mixed_inputs).all() == expected
    assert (copy_labels == mixed_labels).all() == expected



"""Exceptions which allow users to finish pipes abruptly.

Usually shouldn't be used except for the above.

"""


class EpochsException(Exception):
    """Special exception caught by `torchtraining.epochs.EpochsBase` objects.

    User should throw this exception if he wants to stop current iteration immediately
    and proceed with the rest of program.

    Should be used with cautious.

    Currently used by `torchtraining.callbacks.EarlyStopping`, `torchtraining.callbacks.TimeStopping`
    and `torchtraining.callbacks.TerminateOnNan`.
    """

    pass


class IterationsException(Exception):
    """Special exception caught by `torchtraining.iterations.IterationsBase` objects.

    User should throw this exception if he wants to stop current iteration immediately
    and proceed with the rest of program.

    Should be used with cautious.

    """

    pass


class EarlyStopping(EpochsException):
    """EarlyStopping special exception"""

    pass


class TimeStopping(EpochsException):
    """TimeStopping special exception"""

    pass


class TerminateOnNan(EpochsException):
    """TerminateOnNaN special exception"""

    pass



"""
This module has no object oriented counterpart and should be used
in a functional fashion inside `forward` of `torchtraining.steps` (as the only one currently).

See `examples` for specific user cases

"""

import typing

import torch


def mixup(
    inputs: torch.Tensor, targets: torch.Tensor, gamma: float
) -> typing.Tuple[torch.Tensor, torch.Tensor]:
    """Perform per-batch mixup on images.

    See `mixup: Beyond Empirical Risk Minimization <https://arxiv.org/abs/1710.09412>`__.
    for explanation of the method.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                images, labels = sample
                images, labels = tt.functional.inputs.mixup(images, labels)
                # Calculate what you want below, say loss
                ...
                return loss


        step = TrainStep(criterion, device)

    .. note::

        **IMPORTANT**: Examples are modified in-place!

    Arguments
    ---------
    inputs: torch.Tensor
        `torch.Tensor` of shape :math:`(N, *)` and numerical `dtype`.
    labels: torch.Tensor
        `torch.Tensor` of shape :math:`(N, *)` and numerical `dtype`.
    gamma: float
        Level of mixing between inputs and labels. The smaller the value,
        the more "concrete" examples are (e.g. for  `0.1` and `cat`, `dog` labels
        it would be `0.9` cat and `0.1` dog).

    Returns
    -------
    Tuple(torch.Tensor, torch.Tensor)
        Inputs and labels after mixup (linear mix with `gamma` strength).

    """
    if inputs.shape[0] != targets.shape[0]:
        raise ValueError(
            "inputs and labels 0 dimension (batch) has to be equal, "
            "got {} for inputs and {} for labels".format(
                inputs.shape[0], targets.shape[0]
            )
        )
    perm = torch.randperm(inputs.shape[0])
    perm_inputs = inputs[perm]
    perm_targets = targets[perm]
    return (
        inputs.mul_(gamma).add_(perm_inputs, alpha=1 - gamma),
        targets.mul_(gamma).add_(perm_targets, alpha=1 - gamma),
    )



import typing

import torch
from rich import progress

from . import _base, exceptions
from .utils import iterations as iterations_utils


@iterations_utils.docs(
    header="Perform `step` (`train` or `eval`) until `data` is exhausted",
    body="Provided `module` will be passed to every `step`.",
)
class Iteration(_base.Iteration):
    def __init__(
        self,
        step: typing.Any,
        module: torch.nn.Module,
        data: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
        train: bool,
        log: typing.Union[int, str] = "NONE",
        *args,
        **kwargs,
    ):
        super().__init__()

        self.step = step
        self.module = module
        self.data = data
        self.train = train

        self.log = log
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        self.module.train(self.train)
        with torch.set_grad_enabled(self.train):
            yield from super().__call__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        for sample in progress.track(
            self.data, description=" ", *self.args, **self.kwargs
        ):
            yield self.step(self.module, sample, *args, **kwargs)


@iterations_utils.docs(
    header="Perform training step until `data` is exhausted",
    body="Provided `module` will be passed to every `step`.",
)
class Train(Iteration):
    def __init__(
        self,
        step: typing.Any,
        module: torch.nn.Module,
        data: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
        log: typing.Union[int, str] = "NONE",
        *args,
        **kwargs,
    ):
        super().__init__(step, module, data, True, log, *args, **kwargs)


@iterations_utils.docs(
    header="Perform evaluation step until `data` is exhausted",
    body="Provided `module` will be passed to every `step`.",
)
class Eval(Iteration):
    def __init__(
        self,
        step: typing.Any,
        module: torch.nn.Module,
        data: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
        log: typing.Union[int, str] = "NONE",
        *args,
        **kwargs,
    ):
        super().__init__(step, module, data, False, log, *args, **kwargs)


# @utils.iterations.docs(
#     header="Perform evaluation `step`s until `data` is exhausted",
#     body="Provided `module` will be passed to every `step`.",
# )
# class MultiIteration(_base.Iteration):
#     def __init__(
#         self,
#         steps: typing.Iterable[typing.Callable[..., typing.Any]],
#         modules: torch.nn.Module
#         datas: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
#         train: bool,
#         intervals: typing.Optional[typing.Iterable[int]] = None,
#         log: typing.Union[int, str] = "NONE",
#         *args,
#         **kwargs,
#     ):
#         super().__init__()

#         self.steps = steps
#         self.modules = modules
#         self.datas = datas
#         self.train = train

#         if intervals is None:
#             self.intervals = tuple(range(len(steps)))
#         self.intervals = intervals

#         self.log = log
#         self.args = args
#         self.kwargs = kwargs

#     def __call__(self, *args, **kwargs):
#         self.module.train(self.train)
#         with torch.set_grad_enabled(self.train):
#             yield from super().__call__(*args, **kwargs)

#     def forward(self, *args, **kwargs):
#         pass


# @utils.iterations.docs(
#     header="Perform training `step`s until `data` is exhausted",
#     body="Provided `module` will be passed to every `step`.",
# )
# class MultiTrain(MultiIteration):
#     def __init__(
#         self,
#         step: typing.Any,
#         module: torch.nn.Module,
#         data: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
#         log: typing.Union[int, str] = "NONE",
#         *args,
#         **kwargs,
#     ):
#         super().__init__(step, module, data, True, log, *args, **kwargs)


# @utils.iterations.docs(
#     header="Perform evaluation `step`s until `data` is exhausted",
#     body="Provided `module` will be passed to every `step`.",
# )
# class MultiEval(MultiIteration):
#     def __init__(
#         self,
#         step: typing.Any,
#         module: torch.nn.Module,
#         data: typing.Union[torch.utils.data.Dataset, torch.utils.data.DataLoader],
#         log: typing.Union[int, str] = "NONE",
#         *args,
#         **kwargs,
#     ):
#         super().__init__(step, module, data, False, log, *args, **kwargs)



import os
import pathlib
import shutil
import tempfile

import pytest
import torch
import torchtraining as tt
from torch.utils.tensorboard import SummaryWriter


@pytest.mark.parametrize(
    "klass,inputs",
    [
        (tt.callbacks.tensorboard.Scalar, 15),
        (tt.callbacks.tensorboard.Scalar, torch.tensor([15]),),
        (tt.callbacks.tensorboard.Scalars, {"foo": 5, "bar": 10}),
        (
            tt.callbacks.tensorboard.Scalars,
            {"foo": torch.tensor([5]), "bar": torch.tensor(10)},
        ),
        (tt.callbacks.tensorboard.Histogram, torch.randint(0, 15, size=(10, 30))),
        # Below need pillow
        # (tt.callbacks.tensorboard.Image, torch.rand(3, 32, 32)),
        # (tt.callbacks.tensorboard.Images, torch.rand(8, 3, 32, 32)),
        # Below needs moviepy
        # (tt.callbacks.tensorboard.Video, torch.rand(8, 6, 3, 32, 32)),
        (tt.callbacks.tensorboard.Audio, torch.rand(1, 100)),
        (tt.callbacks.tensorboard.Text, "example_text"),
    ],
)
def test_tensorboard(klass, inputs):
    directory = pathlib.Path(tempfile.mkdtemp())
    writer = SummaryWriter(directory)
    tb_writer = klass(writer, "example")
    tb_writer(inputs)
    shutil.rmtree(directory, ignore_errors=True)



import itertools

import pytest
import torch
import torchtraining as tt


class Check(tt.Operation):
    def __init__(self, desired_type):
        super().__init__()
        self.desired_type = desired_type

    def forward(self, data):
        assert data.dtype == self.desired_type


@pytest.mark.parametrize(
    "caster, inputs",
    itertools.product(
        [
            tt.cast.BFloat16() ** Check(torch.bfloat16),
            tt.cast.Bool() ** Check(torch.bool),
            tt.cast.Byte() ** Check(torch.uint8),
            tt.cast.Char() ** Check(torch.int8),
            tt.cast.Double() ** Check(torch.double),
            tt.cast.Float() ** Check(torch.float),
            tt.cast.Half() ** Check(torch.half),
            tt.cast.Int() ** Check(torch.int),
            tt.cast.Long() ** Check(torch.long),
            tt.cast.Short() ** Check(torch.short),
        ],
        [torch.randn(2, 3, 4)],
    ),
)
def test_cast(caster, inputs):
    caster(inputs)



import typing

import torch

from .. import utils


@utils.docs
def regression_of_squares(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.sum,
) -> torch.Tensor:
    return reduction(output - torch.mean(target)) ** 2


@utils.docs
def squares_of_residuals(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.sum,
) -> torch.Tensor:
    return reduction(output - target) ** 2


@utils.docs
def r2(output: torch.Tensor, target: torch.Tensor,) -> torch.Tensor:
    return 1 - squares_of_residuals(output, target) / total_of_squares(target)


@utils.docs
def absolute_error(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.mean,
) -> torch.Tensor:
    return reduction(torch.nn.functional.l1_loss(output, target, reduction="none"))


@utils.docs
def squared_error(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.mean,
) -> torch.Tensor:
    return reduction(torch.nn.functional.mse_loss(output, target, reduction="none"))


@utils.docs
def squared_log_error(
    output: torch.Tensor,
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.mean,
) -> torch.Tensor:
    return reduction((torch.log(1 + target) - torch.log(1 + output)) ** 2)


@utils.docs
def adjusted_r2(output: torch.Tensor, target: torch.Tensor, p: int) -> torch.Tensor:
    numel = output.numel()
    return 1 - (1 - r2(output, target)) * ((numel - 1) / (numel - p - 1))


@utils.docs
def max_error(output: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    return torch.max(torch.abs(output - target))


@utils.docs
def total_of_squares(
    target: torch.Tensor,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.sum,
) -> torch.Tensor:
    return reduction(target - torch.mean(target)) ** 2



from ... import utils


def docs(header, reduction=None):
    def arguments():
        return r"""
Arguments
---------
data: Tuple[torch.Tensor, torch.Tensor]
    Tuple containing `outputs` from neural network and `targets` (ground truths).
    `outputs` should be of shape :math:`(N, *, C-1)`, where :math:`C` is the number of classes.
    Should contain `logits` (unnormalized probabilities) or `probabilities` after
    `softmax` activation or similar.
    `targets` should be of shape :math:`(N, *)` and contain integers in the range
    :math:`[0, C-1]`

        """

    def wrapper(klass):
        docstring = r"""{}

Works for both logits and probabilities of `output` out of the box.

Parameters
----------
        """.format(
            header
        )
        if reduction is not None:
            docstring += utils.docs_general.reduction_parameter(reduction)

        klass.forward.__doc__ = arguments() + utils.docs_general.returned(reduction)
        klass.__doc__ = docstring
        return klass

    return wrapper



from . import docs_general


def docstring(header, body):
    def gradient():
        return r"""
        gradient : bool
            Whether to turn gradient on/off (for training/evaluation respectively).
        """

    def device():
        return r"""
        device : torch.device
            Device to which tensors could be casted. Available in `forward` as
            `self.device`
        """

    def wrapper(klass):
        docstring = r"""{}.

        {}

        Parameters
        ----------
        criterion : typing.Callable
            Criterion to use to get loss value. Available in `forward` as `self.criterion`
            attribute.
        """.format(
            header, body
        )

        if docs_general.is_base(klass):
            docstring += gradient()

        klass.__doc__ = docstring + device()
        return klass

    return wrapper



import typing

import torch


def dimensions_number_check(output: torch.Tensor, target: torch.Tensor) -> None:
    """
    Validate output and target number of dimensions.

    Output should have one dimension more than target as those are probabilities
    or logits.

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    """
    if len(output.shape) - 1 != len(target.shape):
        raise ValueError(
            "Got {} number of dimensions for output and {} for target. "
            "Output should have one dimension more than target.".format(
                len(output.shape), len(target.shape),
            )
        )


def shape_check(output: torch.Tensor, target: torch.Tensor) -> None:
    """
    Validate output and target shapes (except last dimension).

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    """
    output_shape = list(output.shape)
    output_shape.pop(1)
    if output_shape != list(target.shape):
        raise ValueError(
            "Got {} shape for output and {} for target. "
            "Shape should be equal except for 1 dimension of output (specifying classes).".format(
                output.shape, target.shape
            )
        )


def check(output: torch.Tensor, target: torch.Tensor) -> None:
    """
    Run all multiclass checks.

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    """
    dimensions_number_check(output, target)
    shape_check(output, target)


def categorical(
    output: torch.Tensor, target: torch.Tensor
) -> typing.Tuple[torch.Tensor, torch.Tensor]:
    """
    Transform output and target to categorical.

    Additionally, `target` will be transformed to `long` for faster comparison
    (as it is already a categorical).

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    Returns
    -------
    typing.Tuple[torch.Tensor, torch.Tensor]:
        Categorical `output` and `target`.

    """
    return torch.argmax(output, dim=1).long(), target.long()


def one_hot(
    output: torch.Tensor, target: torch.Tensor, num_classes: int = -1
) -> typing.Tuple[torch.Tensor, torch.Tensor]:
    """
    Transform output and target to one hot encoding.

    Both will be transformed to categorical beforehand

    Parameters
    ----------
    output : torch.Tensor
        Usually neural network output
    target : torch.Tensor
        Usually desired target

    Returns
    -------
    typing.Tuple[torch.Tensor, torch.Tensor]:
        One hot encoded `output` and `target`.

    """
    categorical_output, categorical_target = categorical(output, target)
    return (
        torch.nn.functional.one_hot(categorical_output, num_classes),
        torch.nn.functional.one_hot(categorical_target, num_classes),
    )



import typing

import torch

from .. import utils


# assert 2 size
@utils.docs
def cosine(
    output: torch.Tensor,
    target: torch.Tensor,
    dim: int = 1,
    eps: float = 1e-08,
    reduction=torch.mean,
) -> torch.Tensor:
    return reduction(
        1 - torch.nn.functional.cosine_similarity(output, target, dim, eps)
    )


@utils.docs
def pairwise(
    output: torch.Tensor,
    target: torch.Tensor,
    p: float = 2.0,
    eps: float = 1e-06,
    reduction: typing.Callable[[torch.Tensor,], torch.Tensor,] = torch.mean,
) -> torch.Tensor:
    return reduction(torch.nn.functional.pairwise_distance(output, target, p, eps))



import torch


class Classifier(torch.nn.Module):
    def __init__(self, in_channels: int, labels: int):
        super().__init__()
        self.convolution = torch.nn.Sequential(
            # [batch 3, 32, 32]
            torch.nn.Conv2d(in_channels, 64, kernel_size=3, padding=1),
            torch.nn.ReLU(),
            torch.nn.BatchNorm2d(64),
            # [batch 64, 8, 8]
        )
        self.linear = torch.nn.Linear(64, labels)

    def forward(self, images):
        x = self.convolution(images)
        pooled = torch.nn.functional.adaptive_avg_pool2d(x, 1).reshape(
            images.shape[0], -1
        )
        predictions = self.linear(pooled)
        return predictions



"""Cast objects (usually `torch.Tensor`) to specific device.

PyTorch's defaults (as of `1.6.0` CPU and GPU) are forwarded for easier usage.

If you wish to use other (non-standard) devices (like TPUs), please use
`Device` class explicitly, for example (TPU)::


    import torch_xla.core.xla_model as xm

    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            # Generate loss and other necessary items
            ...
            return loss, predictions, targets


    step = TrainStep(criterion, device)
    # Select `loss` and perform backpropagation
    step ** tt.Select(loss=0) ** tt.device.Device(xm.xla_device())

.. note::

    **IMPORTANT**: Usually users should use `torchtraining.device.CPU()` cast
    after `iteration` and before `accumulation` in order not to pollute
    GPUs memory.


"""

import importlib

import torch

from ._base import Operation


class CPU(Operation):
    """Cast `object` (usually `torch.Tensor`) to `cpu`.

    .. note::

        **IMPORTANT**: This object should be used most often from this package
        in order to save `GPU`/`TPU` memory.
        See example below

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss, accuracy


        step = TrainStep(criterion, device)
        iteration = tt.iterations.Train(step, module, dataloader)

        # Cast to CPU in order not to inflate GPU memory with `list`
        # You should usually use this OP before accumulation
        iteration ** tt.Select(
            accuracy=1
        ) ** tt.device.CPU() ** tt.accumulators.List() ** tt.callbacks.Logger("Accuracy")

    Parameters
    ----------
    memory_format: torch.memory_format, optional
        The desired memory format of returned Tensor. Default: torch.preserve_format.
        Default: `torch.preserve_format`

    """

    def __init__(self, memory_format=torch.preserve_format):
        super().__init__()
        self.memory_format = memory_format

    def forward(self, data):
        return data.cpu(memory_format=self.memory_format)


class CUDA(Operation):
    """Cast `object` (usually `torch.Tensor`) to cuda enabled device.

    .. note::

        **IMPORTANT**: This object **USUALLY SHOULDN'T BE USED** as it
        **usually** pointlessly pollutes GPU memory.


    Parameters
    ----------
    device: torch.device | int, optional
        Device index to select. It’s a no-op if this argument is a negative integer or None.
        Default: `None`
    non_blocking: bool, optional
        If True and this copy is between CPU and GPU, the copy may occur asynchronously
        with respect to the host. For other cases, this argument has no effect.
        Default: `False`
    memory_format: torch.memory_format, optional
        The desired memory format of returned Tensor. Default: torch.preserve_format.
        Default: `torch.preserve_format`

    """

    def __init__(
        self, device=None, non_blocking=False, memory_format=torch.preserve_format
    ):
        super().__init__()
        self.device = device
        self.non_blocking = non_blocking
        self.memory_format = memory_format

    def forward(self, data):
        return data.cuda(
            self.device, self.non_blocking, memory_format=self.memory_format,
        )


class Device(Operation):
    """Cast `object` to any device (for example `TPU` with `torch_xla` package).

    .. note::

        **IMPORTANT**: This object **USUALLY SHOULDN'T BE USED** as it
        **usually** pointlessly pollutes device memory (unless it's CPU,
        in such case simply use `torchtraining.device.CPU()`).

    See `example` at the beginning of this section.

    Parameters
    ----------
    device: torch.device | Any
        Anything which can be used with `torch.Tensor.to` to cast onto
        specified device.

    """

    def __init__(self, device):
        super().__init__()
        self.device = device

    def forward(self, data):
        return data.to(self.device)



"""Module providing functions (like metrics, losses) directly usable.

.. note::

    **IMPORTANT**: This module should be rarely used and non-functional
    counterparts should be preferred.


For example this::

    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            images, labels = sample
            # Assume predictions are obtained
            ...
            return loss, predictions, labels


    train_step = TrainStep(criterion, device)


    iteration = (
        tt.iterations.Train(train_step, ...)
        ** tt.Select(predictions=1, labels=2)
        ** tt.metrics.classification.binary.Accuracy()
        ** tt.accumulators.Mean()
        ** tt.callbacks.Log("Accuracy")
    )

Should be preferred instead of this (notice `accuracy` calculation in `step`)::


    class TrainStep(tt.steps.Train):
        def forward(self, module, sample):
            images, labels = sample
            # Assume predictions are obtained
            ...
            accuracy = tt.functional.metrics.classification.binary(predictions, labels)
            return loss, accuracy


    train_step = TrainStep(criterion, device)

    iteration = (
        tt.iterations.Train(train_step, ...)
        ** tt.Select(accuracy=1)
        ** tt.accumulators.Mean()
        ** tt.callbacks.Log("Accuracy")
    )

Second approach has the following shortcomings:

    * calculation of metrics is mixed with what your network actually does with inputs and what it produces
    * Step's' `forward` function has more limited usage. If user wants to calculate other metrics they have to change `step` manually instead of simply adding another `**` pipe


.. note::

    **IMPORTANT**: Only reasonable modules to use (currently) in `functional`
    manner (ha, irony), should be `torchtraining.functional.inputs` and
    `torchtraining.functional.data`.


"""

from . import data, inputs, loss, metrics



import torchvision
import itertools

import torch

import torchtraining as tt


def test_random_split():
    dataset = torchvision.datasets.FakeData()
    train, validation, test = tt.functional.data.random_split(dataset, 0.8, 0.1, 0.1)

    assert len(train) == 800
    assert len(validation) == 100
    assert len(test) == 100



import torch
import torchtraining as tt


class Step(tt.steps.Step):
    def forward(self, module, batch):
        images, labels = batch
        images, labels = images.to(self.device), labels.to(self.device)

        predictions = module(images)
        loss = self.criterion(predictions, labels)

        return loss, predictions, labels


def train(optimizer, criterion, device):
    return (
        Step(criterion, gradient=True, device=device)
        ** tt.Select(loss=0)
        ** tt.pytorch.ZeroGrad(optimizer)
        ** tt.pytorch.Backward()
        ** tt.pytorch.Optimize(optimizer)
    )


def eval(criterion, device):
    return Step(criterion, gradient=False, device=device)



"""Core pytorch operations regarding optimization (optimize, schedule) are placed in general tests."""
import pytest
import torch
import torchtraining.pytorch as P


def test_backward():
    backward = P.Backward()
    x = torch.randn(10, requires_grad=True)
    y = x ** 2
    backward(y.sum())
    assert x.grad is not None



from . import docs_general


def docs(header, body):
    def params(klass):
        if "Multi" in klass.__name__:
            return """Parameters
        ----------
        step: torchtraining.steps.Step
            Single step to run. Usually subclass of `torchtraining.steps.Step`, but could be
            any `Callable` taking `module` and `data` arguments and returning anything.
        module : torch.nn.Module
            Torch module (or modules) passed to `step` during each call.
        data : [torch.utils.data.Dataset | torch.utils.data.DataLoader]
            Iterable object (usually data or dataloader) yielding data passed
            to `step`."""

        return """Parameters
        ----------
        step: torchtraining.steps.Step
            Single step to run. Usually subclass of `torchtraining.steps.Step`, but could be
            any `Callable` taking `module` and `data` arguments and returning anything.
        module : torch.nn.Module
            Torch module (or modules) passed to `step` during each call.
        data : [torch.utils.data.Dataset | torch.utils.data.DataLoader]
            Iterable object (usually data or dataloader) yielding data passed
            to `step`."""

    def train():
        return r"""
        train: bool
            Whether `module` should be in training state (`module.train()`)
            with enabled gradient or in evaluation mode (`module.eval()`) with
            disabled gradient
        """

    def log():
        return r"""
        log : str | int, optional
            Severity level for logging object's actions.
            Available levels of logging:
                * NONE          0
                * TRACE 	5
                * DEBUG 	10
                * INFO 	        20
                * SUCCESS 	25
                * WARNING 	30
                * ERROR 	40
                * CRITICAL 	50
            Default: `NONE` (no logging, `0` priority)

        """

    def wrapper(klass):
        docstring = r"""{}.

        {}

        {}

        """.format(
            header, body, params(klass)
        )
        if docs_general.is_base(klass):
            docstring += train()

        klass.__doc__ = docstring + log()
        return klass

    return wrapper



def returned(reduction):
    if reduction is not None:
        return r"""
Returns
-------
torch.Tensor
    If `reduction` is left as default {} is taken and single value returned.
    Otherwise whatever `reduction` returns.

            """.format(
            reduction
        )
    return r"""
Returns
-------
torch.Tensor
    Scalar `tensor`

        """


def reduction_parameter(reduction):
    if reduction == "mean":
        return r"""
reduction: Callable(torch.Tensor) -> Any, optional
    One argument callable getting `torch.Tensor` and returning argument
    after specified reduction.
    Default: `torch.mean` (mean across batch, user can use `torchtraining.savers.Mean`
    to get mean across iterations/epochs).
"""

    if reduction == "sum":
        return r"""
reduction: Callable, optional
    One argument callable getting `torch.Tensor` and returning `torch.Tensor`.
    Default: `torch.sum` (sum of all elements, user can use `torchtraining.savers.Sum`
    to get sum across iterations/epochs).
"""

    raise ValueError(
        """reduction argument can be one of ["sum", "mean"], got {}""".format(reduction)
    )



"""Cast tensors in a functional fashion.

Users can use this module to cast `step` outputs to desired type or
to lower precision in order to save memory (though it shouldn't be needed.)

"""

import abc

import torch

from ._base import Operation


def _docstring(klass) -> str:
    klass.__doc__ = """Cast `torch.Tensor` instance to {cast}.

.. note::

    **IMPORTANT**: Only `torch.Tensor` can be passed as `memory_format`
    is specified during casting.


Returns
-------
{cast}
    Casted `data`

""".format(
        cast=klass.__name__
    )
    return klass


def _forward_docstring(function):
    function.__doc__ = """
    Arguments
    ---------
    data: torch.Tensor
        Tensor to be casted
    """
    return function


class _Cast(Operation):
    """Shared base class amongst most casting operations."""

    def __init__(self, memory_format=torch.preserve_format):
        super().__init__()
        self.memory_format = memory_format

    @abc.abstractmethod
    def forward(self, data):
        pass


@_docstring
class BFloat16(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.bfloat16(memory_format=self.memory_format)


@_docstring
class Bool(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.bool(memory_format=self.memory_format)


@_docstring
class Byte(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.byte(memory_format=self.memory_format)


@_docstring
class Char(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.char(memory_format=self.memory_format)


@_docstring
class Double(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.double(memory_format=self.memory_format)


@_docstring
class Float(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.float(memory_format=self.memory_format)


@_docstring
class Half(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.half(memory_format=self.memory_format)


@_docstring
class Int(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.int(memory_format=self.memory_format)


@_docstring
class Long(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.long(memory_format=self.memory_format)


@_docstring
class Short(_Cast):
    @_forward_docstring
    def forward(self, data):
        return data.short(memory_format=self.memory_format)


@_docstring
class Item(Operation):
    @_forward_docstring
    def forward(self, data):
        return data.item()


@_docstring
class Numpy(Operation):
    @_forward_docstring
    def forward(self, data):
        return data.numpy()


@_docstring
class List(Operation):
    @_forward_docstring
    def forward(self, data):
        return data.to_list()


@_docstring
class MKLDNN(Operation):
    @_forward_docstring
    def forward(self, data):
        return data.to_mkldnn()


class Sparse(Operation):
    """Cast `torch.Tensor` to sparse format.

    Parameters
    ----------
    sparse_dims: int, optional
        The number of sparse dimensions to include in the new sparse tensor.
        Default: `None`.

    """

    def __init__(self, sparse_dims=None):
        super().__init__()
        self.sparse_dims = sparse_dims

    @_forward_docstring
    def forward(self, data):
        return data.to_sparse(sparse_dims=self.sparse_dims)


# As another tensor
class As(Operation):
    """Cast `torch.Tensor` to the same type as `other`.

    Parameters
    ----------
    other: torch.Tensor
        Tensor according to which incoming tensor will be casted.

    """

    def __init__(self, other):
        super().__init__()
        self.other = other

    @_forward_docstring
    def forward(self, data):
        return data.type_as(self.other)


###############################################################################
#
#                               TYPE ALIASES
#
###############################################################################

UInt8 = Byte
Int8 = Char
Int16 = Short
Int32 = Int
Int64 = Long
Float16 = Half
Float32 = Float
Float64 = Double



import itertools

import torch

import pytest
import torchtraining as tt


@pytest.mark.parametrize(
    "cast,inputs",
    itertools.product(
        (
            (tt.device.CPU(), torch.device("cpu")),
            (tt.device.Device(device=torch.device("cpu")), torch.device("cpu")),
        ),
        (torch.randn(4, 5), torch.randn(5)),
    ),
)
def test_cast(cast, inputs):
    caster, desired = cast
    assert caster(inputs).device == desired



import functools
import os

import setuptools

exec(open("torchtraining/_version.py").read())  # get __version__
exec(open("torchtraining/_name.py").read())  # get _name

extras = {
    "horovod": ["horovod[pytorch]"],
    # Loggers
    "tensorboard": [
        "tensorboard>=2.3.0",
        "matplotlib>=3.3.1",  # add_figure
        "pillow>=7.2.0",  # add_image, add_images
        "moviepy>=1.0.3",  # add_video
    ],
    "neptune": ["neptune-client>=0.4.119"],
    "comet": ["comet-ml>=3.1.17"],
}


extras["callbacks"] = extras["tensorboard"] + extras["neptune"] + extras["comet"]
extras["accelerators"] = extras["horovod"]

extras["all"] = list(
    functools.reduce(
        lambda requirements1, requirements2: requirements1 + requirements2,
        extras.values(),
    )
)

extras["tests"] = extras["callbacks"]

setuptools.setup(
    name=_name,
    version=__version__,
    license="MIT",
    author="Szymon Maszke",
    author_email="szymon.maszke@protonmail.com",
    description="Functional & flexible neural network training with PyTorch.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/torchtraining",
    packages=setuptools.find_packages(),
    install_requires=["torch>=1.3.0", "loguru>=0.5.1", "rich>=2.3.0", "PyYAML>=5.3.1"],
    extras_require=extras,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Website": "https://szymonmaszke.github.io/torchtraining",
        "Documentation": "https://szymonmaszke.github.io/torchtraining/#torchtraining",
        "Issues": "https://github.com/szymonmaszke/torchtraining/issues?q=is%3Aissue+is%3Aopen+sort%3Aupdated-desc",
    },
    keywords="pytorch train functional flexible research fit epoch step simple fast",
)



"""This module allows user to run multiple `iterations` in consecutive order,
which constitutes a single epoch.

Currently, only default `Epoch` is provided.

"""

import abc
import contextlib
import typing

import loguru

from . import _base, exceptions


class Epoch(_base.Epoch):
    """
    Loop over specified `iterations` until `epochs` number is reached.

    See `Introduction tutorial<>__` for examples.

    .. note::

        **IMPORTANT**: This module is one of core features
        so be sure to understand how it works.

    .. note::

        **IMPORTANT**: Usually this is the top most class object
        used in training (composing multiple `iterations`).


    Parameters
    ----------
    *iterations : Iterable[torchtraining.iterations.Iteration]
        Iterations to be run one after another during single loop pass.
    epochs : int
        **Keyword only argument** specifying number of epochs to run.
    log : str | int, optional
        Severity level for logging object's actions.
        Available levels of logging:
            * NONE      0
            * TRACE 	5
            * DEBUG 	10
            * INFO 	20
            * SUCCESS 	25
            * WARNING 	30
            * ERROR 	40
            * CRITICAL 	50
        Default: `NONE` (no logging, `0` priority)

    """

    def __init__(self, *iterations, epochs: int, log: typing.Union[str, int] = "NONE"):
        super().__init__()
        self.iterations = iterations
        self.epochs = epochs
        self.log = log

    def __len__(self):
        return self.epochs

    def forward(self, *args, **kwargs):
        for index in range(self.epochs):
            loguru.logger.log(self.log, "Starting epoch {}.".format(index))
            for iteration in self.iterations:
                yield from iteration(*args, **kwargs)

        loguru.logger.log(self.log, "Finished epochs.")



"""Root module of `torchtraining` containing common operations.

.. note::

    **IMPORTANT**: This module is one of the most important and
    is used in almost any DL task so be sure to understand it!


Operations in this module can be used to:

    * control pipeline flow
    * select output from `step`s
    * send data to multiple operations

See below for more info.

"""

import typing

import loguru
import torch
import yaml

from . import (accelerators, accumulators, callbacks, cast, device, epochs,
               exceptions, functional, iterations, loss, metrics, pytorch,
               quantization, steps)
from ._base import (Accumulator, Epoch, GeneratorProducer, Iteration,
                    Operation, ReturnProducer, Step)
from ._version import __version__

loguru.logger.level("NONE", no=0)


class Select(Operation):
    """Select output item(s) returned from `step` or `iteration` objects.

    Allows users to focus on specific part of output and pipe specified
    values to other operations (like `metrics`, `loggers` etc.).

    .. note::

        **IMPORTANT**: This operation is run in almost any case
        so be sure to understand how it works.


    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                # Generate loss and other necessary items
                ...
                return loss, predictions, targets


        step = TrainStep(criterion, device)
        # Select `loss` and perform backpropagation
        # Only single value will be forward to backward from
        # (loss, predictions, targets) tuple
        step ** tt.Select(loss=0) ** tt.pytorch.Backward()

    .. note::

        Name of keyword argument can be arbitrary but
        **you are really encouraged** to name it like the variable
        returned from `step` (or at least make it understandable to others).


    Parameters
    ----------
    output_selection : **output_selection
        `name`: output_index mapping selecting which element from step returned `tuple`
        to choose. `name` can be arbitrary, but should be named like the variable
        returned from `step`. See example above.


    Returns
    -------
    Any | List[Any]
        If single int is passed `output_selection` return single element from `Iterable`.
        Otherwise returns chosen elements as `list`

    """

    def __init__(self, **output_selection: int):
        if len(output_selection) < 1:
            raise ValueError(
                "{}: At least one output index has to be specified, got {} output indices.".format(
                    self, len(output_selection)
                )
            )

        super().__init__()
        self.output_selection = output_selection

        if len(self.output_selection) == 1:
            self._selection_method = lambda data: data[
                list(self.output_selection.values())[0]
            ]
        else:
            self._selection_method = lambda data: [
                data[index] for index in list(self.output_selection.values())
            ]

    def forward(self, data: typing.Iterable[typing.Any]) -> typing.Any:
        """
        Arguments
        ---------
        data: Iterable[Any]
            Iterable (e.g. `tuple`, `list`) with any elements.
        """
        selected = self._selection_method(data)
        return selected

    def __str__(self) -> str:
        return yaml.dump({super().__str__(): self.output_selection})


class Split(Operation):
    """Split operation with data to multiple components.

    .. note::

        **IMPORTANT**: This operation is run in almost any case
        so be sure to understand how it works.


    Useful when users wish to use calculated result in multiple places.
    Example calculating metrics and logging them in multiple places::


        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                # Generate loss and other necessary items
                ...
                # Assume binary classification
                return loss, logits, targets


        step = TrainStep(criterion, device)

        # Push (logits, targets) to Precision and Recall
        # and log those values after calculating metrics
        step ** tt.Select(logits=1, targets=2) ** tt.Split(
            tt.metrics.classification.binary.Precision() ** tt.callbacks.Logger("Precision"),
            tt.metrics.classification.binary.Recall() ** tt.callbacks.Logger("Recall"),
        )


    Parameters
    ----------
    *operations: Callable(data) -> Any
        Operations to which results will be passed.
    return_modified: bool, optional
        Return outputs from `operations` as a `list` if `True`. If `False`, returns
        original `data` passed into `Split`. Default: `False`

    Returns
    -------
    data | List[modified data]
        Returns `data` passed originally or `list` containing modified data
        returned from `operations`.

    """

    def __init__(
        self,
        *operations: typing.Callable[[typing.Any,], typing.Any],
        return_modified: bool = False
    ):
        if not operations:
            raise ValueError("Split requires at least one operation to pass data into.")
        super().__init__()
        self.operations = operations
        self.return_modified = return_modified

    def forward(
        self, data: typing.Any
    ) -> typing.Union[typing.Any, typing.List[typing.Any]]:
        """
        Arguments
        ---------
        data: Any
            Data which will be passed to provided `operations`.
        """
        processed_data = []
        for operation in self.operations:
            result = operation(data)
            if self.return_modified:
                processed_data.append(result)
        if self.return_modified:
            return processed_data
        return data

    def __str__(self) -> str:
        return yaml.dump({super().__str__(): self.operations})


class OnSplittedTensor(Operation):
    """Split tensor along dimension and apply operation on each element.

    By default, `torch.Tensor` will be splitted along batch (`dim=0`)

    .. note::

        **IMPORTANT**: After splitting first dimension is squeezed
        via `torch.squeeze`.


    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                # Dummy step
                images, labels
                return images


        step = TrainStep(criterion, device)

        # Assume summary_writer is instance of torch.utils.tensorboard.SummaryWriter()
        step ** tt.OnSplittedTensor(tt.callbacks.tensorboard.Image(summary_writer))
        # Each image will be saved separately



    Parameters
    ----------
    operation: tt.Operation | Callable(data) -> Any
        Operation which will be applied to each element of `torch.Tensor`.
    dim: int, optional
        Dimension along which `data` `torch.Tensor` will be splitted.
        Default: `0`


    Returns
    -------
    Tuple[torch.Tensor]
        Splitted `data` along `dim` (unmodified by `operation`).

    """

    def __init__(self, operation: Operation, dim: int = 0):
        super().__init__()
        self.operation = operation
        self.dim = dim

    def forward(self, data):
        """
        Arguments
        ---------
        data: torch.Tensor
            Tensor to split and apply `operation` on.
        """
        splitted = tuple(map(torch.squeeze, torch.split(data, 1, self.dim)))
        for tensor in splitted:
            self.operation(splitted)
        return splitted


class Flatten(Operation):
    r"""Flatten arbitrarily nested data.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                module1, module2 = module
                ...
                return ((logits1, targets1), (logits2, targets2), module1, module2)


        step = TrainStep(criterion, device)

        # Tuple (logits1, targets1, logits2, targets2, module1, module2)
        step ** tt.Flatten()

    Parameters
    ----------
    types : Tuple[type], optional
        Types to be considered non-flat. Those will be recursively flattened.
        Default: `(list, tuple)`

    Returns
    -------
    Tuple[samples]
        Single `tuple` with all elements (not being `tuple` or `list`).

    """

    def __init__(self, types: typing.Tuple = (list, tuple)):
        super().__init__()
        self.types = types

    def forward(self, data) -> typing.List[typing.Any]:
        """
        Parameters
        ----------
        data: Iterable[Iterable ... Iterable[Any]]
            Arbitrarily nested data being one of type provided in `types`.
        """
        if not isinstance(data, self.types):
            return data
        return Flatten._flatten(data, self.types)

    @staticmethod
    def _flatten(
        items: typing.Iterable[typing.Any], types: typing.Tuple
    ) -> typing.List[typing.Any]:
        if isinstance(items, tuple):
            items = list(items)

        for index, x in enumerate(items):
            while index < len(items) and isinstance(items[index], types):
                items[index : index + 1] = items[index]
        return items


class If(Operation):
    """Run operation only If `condition` is `True`.

    `condition` can also be a single argument callable, in this case it can be
    dependent on data, see below::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss


        step = TrainStep(criterion, device)
        step ** tt.If(lambda loss: loss ** 10, tt.callbacks.Logger("VERY HIGH LOSS!!!"))

    Parameters
    ----------
    condition: bool | Callable(Any) -> bool
        If boolean value and if `True`, run underlying Operation (or other Callable).
        If Callable, should take data as argument and return decision based on
        that as single `bool`.
    operation: torchtraining.Operation | Callable
        Operation to run if `True`


    Returns
    -------
    Any
        If `true`, returns value from `operation`, otherwise passes original `data`

    """

    def __init__(
        self,
        condition: typing.Union[bool, typing.Callable[[typing.Any], bool,]],
        operation: typing.Callable[[typing.Any,], typing.Any],
    ):
        super().__init__()
        if isinstance(condition, bool):
            self._choice_method = lambda data: condition
        else:
            self._choice_method = lambda data: condition(data)

        self.condition = condition
        self.operation = operation

    def forward(self, data: typing.Any) -> typing.Any:
        """
        Arguments
        ---------
        data: Any
            Anything you want (usually `torch.Tensor` like stuff).
        """
        if self.condition(data):
            return self.operation(data)
        return data

    def __str__(self) -> str:
        if self.condition:
            return str(self.operation)
        return "no-operation"


# Make it dynamic and static
class IfElse(Operation):
    """Run `operation1` only if `condition` is `True`, otherwise run `operation2`.

    `condition` can also be a single argument callable, in this case it can be
    dependent on data, see below::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss


        step = TrainStep(criterion, device)

        step ** tt.IfElse(
            lambda loss: loss ** 10,
            tt.callbacks.Logger("VERY HIGH LOSS!!!"),
            tt.callbacks.Logger("LOSS IS NOT THAT HIGH..."),
        )

    Parameters
    ----------
    condition: bool
        Boolean value. If `true`, run underlying Op (or other Callable).
    operation1: torchtraining.Op | Callable
        Operation or callable getting single argument (`data`) and returning anything.
    operation2: torchtraining.Op | Callable
        Operation or callable getting single argument (`data`) and returning anything.


    Returns
    -------
    Any
        If `true`, returns value from `operation1`, otherwise from `operation2`.

    """

    def __init__(
        self,
        condition: bool,
        operation1: typing.Callable[[typing.Any,], typing.Any],
        operation2: typing.Callable[[typing.Any,], typing.Any],
    ):
        super().__init__()
        self.condition = condition
        self.operation1 = operation1
        self.operation2 = operation2

    def forward(self, data: typing.Any) -> typing.Any:
        """
        Arguments
        ---------
        data: Any
            Anything you want (usually `torch.Tensor` like stuff).
        """
        if self.condition:
            return self.operation1(data)
        return self.operation2(data)

    def __str__(self) -> str:
        if self.condition:
            return str(self.operation1)
        return str(self.operation2)


class ToAll(Operation):
    r"""Apply operation to each element of sample.**

    .. note::

        If you want to apply operation to all nested elements (e.g. in nested `tuple`),
        please use `torchtraining.Flatten` object first.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return loss


        step = TrainStep(criterion, device)

        step ** tt.If(
            lambda loss: loss ** 10,
            tt.callbacks.Logger("VERY HIGH LOSS!!!"),
            tt.callbacks.Logger("LOSS IS NOT THAT HIGH..."),
        )


    Parameters
    ----------
    operation: Callable
        Pipe to apply to each element of sample.

    Returns
    -------
    Tuple[operation(subsample)]
        Tuple consisting of subsamples with operation applied.

    """

    def __init__(self, operation: typing.Callable):
        super().__init__()
        self.operation = operation

    def forward(self, sample):
        """
        Arguments
        ---------
        data: Any
            Anything you want (usually `torch.Tensor` like stuff).
        """
        return tuple(self.operation(subsample) for subsample in sample)


class Lambda(Operation):
    """Run user specified operation on `data`.

    Example::

        class TrainStep(tt.steps.Train):
            def forward(self, module, sample):
                ...
                return accuracy


        step = TrainStep(criterion, device)

        # If you want to get that SOTA badly, we got ya covered
        step ** tt.Lambda(lambda accuracy: accuracy * 2)

    Parameters
    ----------
    operation: Callable(Any) -> Any
        Single argument callable getting data and returning some value.
    name: str, optional
        `string` representation of this operation (if any).
        Default: `torchtraining.metrics.Lambda`


    Returns
    -------
    Any
        Value returned from `operation`

    """

    def __init__(
        self,
        operation: typing.Callable[[typing.Any,], typing.Any],
        name: str = "torchtraining.Lambda",
    ):
        super().__init__()
        self.operation = operation
        self.name = name

    def __str__(self) -> str:
        return self.name

    def forward(self, data: typing.Any) -> typing.Any:
        """
        Arguments
        ---------
        data: Any
            Anything you want (usually `torch.Tensor` like stuff).
        """
        return self.operation(data)
